from contextlib import contextmanager
from pyxk.httpx._client import Client


def request(
    method, url, *, params=None, content=None, data=None, files=None, json=None, headers=None, cookies=None, auth=None, proxies=None, timeout=5, follow_redirects=True, verify=True, cert=None, trust_env=True
):
    with Client(
        cookies=cookies, proxies=proxies, cert=cert, verify=verify, timeout=timeout, trust_env=trust_env
    ) as client:
        return client.request(
            method=method, url=url, content=content, data=data, files=files, json=json, params=params, headers=headers, auth=auth, follow_redirects=follow_redirects
        )


def get(
    url, *, params=None, headers=None, cookies=None, auth=None, proxies=None, follow_redirects=True, cert=None, verify=True, timeout=5, trust_env=True
):
    return request(
        "GET", url, params=params, headers=headers, cookies=cookies, auth=auth, proxies=proxies, follow_redirects=follow_redirects, cert=cert, verify=verify, timeout=timeout, trust_env=trust_env
    )


def options(
    url, *, params=None, headers=None, cookies=None, auth=None, proxies=None, follow_redirects=True, cert=None, verify=True, timeout=5, trust_env=True
):
    return request(
        "OPTIONS", url, params=params, headers=headers, cookies=cookies, auth=auth, proxies=proxies, follow_redirects=follow_redirects, cert=cert, verify=verify, timeout=timeout, trust_env=trust_env
    )


def head(
    url, *, params=None, headers=None, cookies=None, auth=None, proxies=None, follow_redirects=False, cert=None, verify=True, timeout=5, trust_env=True
):
    return request(
        "HEAD", url, params=params, headers=headers, cookies=cookies, auth=auth, proxies=proxies, follow_redirects=follow_redirects, cert=cert, verify=verify, timeout=timeout, trust_env=trust_env
    )


def post(
    url, *, content=None, data=None, files=None, json=None, params=None, headers=None, cookies=None, auth=None, proxies=None, follow_redirects=True, cert=None, verify=True, timeout=5, trust_env=True
):
    return request(
        "POST", url, content=content, data=data, files=files, json=json, params=params, headers=headers, cookies=cookies, auth=auth, proxies=proxies, follow_redirects=follow_redirects, cert=cert, verify=verify, timeout=timeout, trust_env=trust_env
    )


def put(
    url, *, content=None, data=None, files=None, json=None, params=None, headers=None, cookies=None, auth=None, proxies=None, follow_redirects=True, cert=None, verify=True, timeout=5, trust_env=True
):
    return request(
        "PUT", url, content=content, data=data, files=files, json=json, params=params, headers=headers, cookies=cookies, auth=auth, proxies=proxies, follow_redirects=follow_redirects, cert=cert, verify=verify, timeout=timeout, trust_env=trust_env
    )


def patch(
    url, *, content=None, data=None, files=None, json=None, params=None, headers=None, cookies=None, auth=None, proxies=None, follow_redirects=True, cert=None, verify=True, timeout=5, trust_env=True
):
    return request(
        "PATCH", url, content=content, data=data, files=files, json=json, params=params, headers=headers, cookies=cookies, auth=auth, proxies=proxies, follow_redirects=follow_redirects, cert=cert, verify=verify, timeout=timeout, trust_env=trust_env
    )


def delete(
    url, *, params=None, headers=None, cookies=None, auth=None, proxies=None, follow_redirects=True, cert=None, verify=True, timeout=5, trust_env=True
):
    return request(
        "DELETE", url, params=params, headers=headers, cookies=cookies, auth=auth, proxies=proxies, follow_redirects=follow_redirects, cert=cert, verify=verify, timeout=timeout, trust_env=trust_env
    )


@contextmanager
def stream(
    method, url, *, params=None, content=None, data=None, files=None, json=None, headers=None, cookies=None, auth=None, proxies=None, timeout=5, follow_redirects=True, verify=True, cert=None, trust_env=True
):
    with Client(
        cookies=cookies, proxies=proxies, cert=cert, verify=verify, timeout=timeout, trust_env=trust_env
    ) as client:
        with client.stream(
            method=method, url=url, content=content, data=data, files=files, json=json, params=params, headers=headers, auth=auth, follow_redirects=follow_redirects
        ) as response:
            yield response


def wget(
    url, method="GET", *, params=None, content=None, data=None, files=None, json=None, headers=None, cookies=None, auth=None, proxies=None, timeout=5, follow_redirects=True, verify=True, cert=None, trust_env=True, output=None, resume=False
):
    with Client(
        cookies=cookies, proxies=proxies, cert=cert, verify=verify, timeout=timeout, trust_env=trust_env
    ) as client:
        return client.wget(
            method=method, url=url, content=content, data=data, files=files, json=json, params=params, headers=headers, auth=auth, follow_redirects=follow_redirects, output=output, resume=resume
        )



from pyxk.requests.sessions import Session


def request(
    method, url, *, params=None, data=None, headers=None, cookies=None, files=None, auth=None, timeout=5, allow_redirects=True, proxies=None, hooks=None, stream=None, verify=None, cert=None, json=None
):
    with Session() as session:
        return session.request(
            method=method, url=url, params=params, data=data, headers=headers, cookies=cookies, files=files, auth=auth, timeout=timeout, allow_redirects=allow_redirects, proxies=proxies, hooks=hooks, stream=stream, verify=verify, cert=cert, json=json
        )


def get(
    url, *, params=None, data=None, headers=None, cookies=None, files=None, auth=None, timeout=5, allow_redirects=True, proxies=None, hooks=None, stream=None, verify=None, cert=None, json=None
):
    return request(
        method="GET", url=url, params=params, data=data, headers=headers, cookies=cookies, files=files, auth=auth, timeout=timeout, allow_redirects=allow_redirects, proxies=proxies, hooks=hooks, stream=stream, verify=verify, cert=cert, json=json
    )


def post(
    url, *, params=None, data=None, headers=None, cookies=None, files=None, auth=None, timeout=5, allow_redirects=True, proxies=None, hooks=None, stream=None, verify=None, cert=None, json=None
):
    return request(
        method="POST", url=url, params=params, data=data, headers=headers, cookies=cookies, files=files, auth=auth, timeout=timeout, allow_redirects=allow_redirects, proxies=proxies, hooks=hooks, stream=stream, verify=verify, cert=cert, json=json
    )


def options(
    url, *, params=None, data=None, headers=None, cookies=None, files=None, auth=None, timeout=5, allow_redirects=True, proxies=None, hooks=None, stream=None, verify=None, cert=None, json=None
):
    return request(
        method="OPTIONS", url=url, params=params, data=data, headers=headers, cookies=cookies, files=files, auth=auth, timeout=timeout, allow_redirects=allow_redirects, proxies=proxies, hooks=hooks, stream=stream, verify=verify, cert=cert, json=json
    )


def head(
    url, *, params=None, data=None, headers=None, cookies=None, files=None, auth=None, timeout=5, allow_redirects=False, proxies=None, hooks=None, stream=None, verify=None, cert=None, json=None
):
    return request(
        method="HEAD", url=url, params=params, data=data, headers=headers, cookies=cookies, files=files, auth=auth, timeout=timeout, allow_redirects=allow_redirects, proxies=proxies, hooks=hooks, stream=stream, verify=verify, cert=cert, json=json
    )


def put(
    url, *, params=None, data=None, headers=None, cookies=None, files=None, auth=None, timeout=5, allow_redirects=True, proxies=None, hooks=None, stream=None, verify=None, cert=None, json=None
):
    return request(
        method="PUT", url=url, params=params, data=data, headers=headers, cookies=cookies, files=files, auth=auth, timeout=timeout, allow_redirects=allow_redirects, proxies=proxies, hooks=hooks, stream=stream, verify=verify, cert=cert, json=json
    )


def patch(
    url, *, params=None, data=None, headers=None, cookies=None, files=None, auth=None, timeout=5, allow_redirects=True, proxies=None, hooks=None, stream=None, verify=None, cert=None, json=None
):
    return request(
        method="PATCH", url=url, params=params, data=data, headers=headers, cookies=cookies, files=files, auth=auth, timeout=timeout, allow_redirects=allow_redirects, proxies=proxies, hooks=hooks, stream=stream, verify=verify, cert=cert, json=json
    )


def delete(
    url, *, params=None, data=None, headers=None, cookies=None, files=None, auth=None, timeout=5, allow_redirects=True, proxies=None, hooks=None, stream=None, verify=None, cert=None, json=None
):
    return request(
        method="DELETE", url=url, params=params, data=data, headers=headers, cookies=cookies, files=files, auth=auth, timeout=timeout, allow_redirects=allow_redirects, proxies=proxies, hooks=hooks, stream=stream, verify=verify, cert=cert, json=json
    )


def wget(
    url, method="GET", *, params=None, data=None, headers=None, cookies=None, files=None, auth=None, timeout=5, allow_redirects=True, proxies=None, hooks=None, verify=None, cert=None, json=None, stream=True, output=None, resume=False
):
    """大文件流式响应下载

    :params: output: 下载文件路径
    :params: resume: 断点续传(default: False)
    """
    session = Session()
    try:
        response = session.wget(
            method=method, url=url, params=params, data=data, headers=headers, cookies=cookies, files=files, auth=auth, timeout=timeout, allow_redirects=allow_redirects, proxies=proxies, hooks=hooks, verify=verify, cert=cert, json=json, stream=stream, output=output, resume=resume
        )
    finally:
        session.close()
    return response



import ast
import click
from pyxk.m3u8._parser import M3U8


@click.command()
@click.argument("uri", metavar="m3u8链接地址", type=str)
@click.option("-o", "--output", help="下载保存文件名称", default=None, type=str)
@click.option("-r", "--rerequest/--no-rerequest", help="重新请求m3u8资源", default=False)
@click.option("--delete/--no-delete", help="下载完成后, 删除m3u8本地资源", default=True)
@click.option("--reserve/--no-reserve", help="下载完成后, 保留segments", default=False)
@click.option("--headers", help="request请求头", default=None, type=str)
@click.option("--verify/--no-verify", help="request SSL验证", default=True)
@click.option("--limit", help="异步下载limit", default=8, type=int)
def main(
    uri, output, rerequest, delete, reserve, headers, verify, limit
):
    m3u8parse_params = {
        "file": output,
        "headers": headers,
        "verify": verify,
        "rerequest": rerequest,
        "delete": delete,
        "reserve": reserve,
        "limit": limit,
    }
    m3u8_obj = M3U8()
    m3u8_obj.load(uri=uri, **m3u8parse_params)



"""
延迟加载实现
"""
from types import ModuleType
from importlib import import_module


class LazyLoader(ModuleType):
    """
    模块 延迟加载
    :params: local_name: 模块引用名称
    :params: module_life_cycle: 模块生命周期( 建议使用全局变量 globals() )
    :params: import_name: 导入模块名称
    """
    def __init__(
        self, local_name, module_life_cycle, import_name=None
    ):
        self._local_name = local_name
        self._module_life_cycle = module_life_cycle
        super().__init__(import_name or local_name)

    def _loader(self):
        module = import_module(self.__name__)
        self._module_life_cycle[self._local_name] = module
        self.__dict__.update(module.__dict__)
        return module

    def __getattr__(self, name):
        module = self._loader()
        return getattr(module, name)

    def __dir__(self):
        module = self._loader()
        return dir(module)



"""
AES 加解密
"""
from pyxk.lazy_loader import LazyLoader

copy = LazyLoader("copy", globals(), "copy")
AES = LazyLoader("AES", globals(), "Crypto.Cipher.AES")
typing = LazyLoader("typing", globals(), "typing")
fmtdata = LazyLoader("fmtdata", globals(), "pyxk.aes._fmtdata")



def no_padding(data, remove=False, pad=b"\x00"):
    """
    NoPadding填充模式
    """
    # 消除 padding 字符
    if remove:
        return data.rstrip(pad)
    remainder = len(data) % AES.block_size or AES.block_size
    data += pad * (AES.block_size - remainder)
    return data


def zero_padding(data, remove=False, pad=b"\x00"):
    """
    ZeroPadding填充模式
    """
    # 消除 padding 字符
    if remove:
        return data.rstrip(pad)
    remainder = len(data) % AES.block_size
    # 不填充
    data += pad * (AES.block_size - remainder)
    return data


PADDING_ALL = {
    "Raw": lambda data, *args, **kwagrs: data,
    "NoPadding": no_padding,
    "ZeroPadding": zero_padding,
}



class Cryptor(fmtdata.FormatData):
    """
    AES加解密
    """

    def __init__(
        self, key, iv=None, mode="CBC", padding="NoPadding", **kwargs
    ):
        self._cipher = None
        self._padding = padding
        self.__padding_fmt()
        super().__init__(key, iv, mode, **kwargs)


    def __padding_fmt(self):
        """
        加解密数据的填充方式
        """
        padding = getattr(self, "_padding", None)
        if padding is None:
            setattr(self, "_padding", "NoPadding")
            return

        if (
            not isinstance(padding, str)
            or padding not in PADDING_ALL
        ):
            raise ValueError(
                f"\033[31m'padding' must exist in the {list(PADDING_ALL)},"
                f" not '{padding}'\033[0m")


    def encrypt(self, plaintext):
        """
        加密
        """
        if isinstance(plaintext, str):
            plaintext = plaintext.encode(self._encode)

        elif not isinstance(plaintext, bytes):
            raise TypeError(
                "\033[31m'plaintext' type must be 'str' or 'bytes',"
                f" not '{type(plaintext).__name__}'\033[0m")

        # 创建 cipher - 加密
        self.__create_cipher()
        padding_func = PADDING_ALL[self.padding]
        return self._cipher.encrypt( padding_func(plaintext) )


    def decrypt(self, ciphertext):
        """
        解密
        """
        if isinstance(ciphertext, str):
            ciphertext = ciphertext.encode(self._encode)

        elif not isinstance(ciphertext, bytes):
            raise TypeError(
                "\033[31m'plaintext' type must be 'str' or 'bytes',"
                f" not '{type(ciphertext).__name__}'\033[0m")

        # 创建 cipher - 解密
        self.__create_cipher()
        padding_func = PADDING_ALL[self.padding]
        return padding_func(self._cipher.decrypt(ciphertext), True)


    def __create_cipher(self):
        """
        创建 cipher
        """
        state = copy.deepcopy(self._state)
        state["key"]  = self.key
        state["mode"] = self.mode
        if self.iv is not None:
            state["iv"] = self._iv
        setattr(self, "_cipher", AES.new(**state))


    @property
    def padding(self):
        if not hasattr(self, "_padding"):
            setattr(self, "_padding", "NoPadding")
        return getattr(self, "_padding")


    @padding.setter
    def padding(self, value):
        setattr(self, "_padding", value)
        self.__padding_fmt()


    def __enter__(self):
        return self


    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            raise
        self._cipher = None



def encrypt(
    key: typing.Union[str, bytes],
    plaintext: typing.Union[str, bytes],
    *,
    mode: typing.Union[int, str]="CBC",
    iv: typing.Union[str, bytes]=None,
    **kwargs
) -> bytes:
    """AES 加密

    :params: key: 加密密钥
    :params: plaintext: 加密明文
    :params: mode: 加密模式
    :params: iv: 加密偏移量(部分加密模式不需要偏移量)
    :params: **kwargs: 可选关键字参数
    """
    with Cryptor(key=key, mode=mode, iv=iv, **kwargs) as _cipher:
        return _cipher.encrypt(plaintext)


def decrypt(
    key: typing.Union[str, bytes],
    ciphertext: typing.Union[str, bytes],
    *,
    mode: typing.Union[int, str]="CBC",
    iv: typing.Union[str, bytes]=None,
    **kwargs
) -> bytes:
    """AES 解密

    :params: key: 解密密钥
    :params: ciphertext: 解密密文
    :params: mode: 解密模式
    :params: iv: 解密偏移量(部分解密模式不需要偏移量)
    :params: **kwargs: 可选关键字参数
    """
    with Cryptor(key=key, mode=mode, iv=iv, **kwargs) as _cipher:
        return _cipher.decrypt(ciphertext)
